"""
Binder classes perform two functions through their format method

- given a query template with %(somevar)s python substition


    class MyClass(object):
        pass

    arg1 = MyClass()
    arg1.customer = 101

    default = MyClass()
    default.customer = 201
    arg2.country = "CAN"

    qry, sub = format("
        select *
        from customer
        where country = %(country)s
        and custid = %(customer)s", arg1, default)

    means that we will be fetching for country=CAN, custid=101

- the query template itself is transformed to a format
  that fits the underlying database's bind variable
  scheme which protects against sql injection attacks.

  For example, assuming an Oracle database (paramstyle="named")

  qry:
      "select * from customer where country = :country and custid = :customer"
  sub:
      {"country":"CAN", "customer" : 101}

  Postgres (paramstyle=""):

  qry:
      "select * from customer where country = :country and custid = :customer"
  sub:
      {"country":"CAN", "customer" : 101}


  a positional database (paramstyle="numeric") (NotImplementedError)
  would instead return

  qry:
      "select * from customer where country = :1 and custid = :2"
  sub:
      ["CAN", 101]


"""

import re


class Binder(object):
    """query template and substitution management - generic
    """

    def __init__(self, *args, **kwds):
        pass

    def format(self, tqry, *args):
        """
        :param tqry: query with optional substitution variables
                     Python style i.e.
                     select * from orders where custid = %(custid)s
        :param *args: zero or more arguments that will be checked
                      left-to-right, argument[<key>], getattr(argument,<key>)
        """

    def __repr__(self):
        msg = "%s paramstyle=%s" % (self.__class__.__name__, self.paramstyle)

        if hasattr(self, "supports"):
            msg += " supports: %s" % (self.supports)

        return msg

    def _case_sensitive(self, key):
        return [key]

    def _case_insensitive(self, key):
        if key == key.upper():
            return [key, key.lower()]
        if key == key.lower():
            return [key, key.upper()]
        return [key]

    _key_expand = _case_sensitive

    def _get_from_args(self, key_in):
        """generic way to look for a key in the arg list"""

        li_key = self._key_expand(key_in)

        for key in li_key:
            for arg in self.li_arg:
                try:
                    got = arg[key]
                    return got
                except (KeyError):
                    try:
                        # try getattr
                        got = getattr(arg, key)
                        return got
                    except AttributeError:
                        continue

                except (AttributeError, TypeError):
                    # no __getitem__, try getattr
                    try:
                        got = getattr(arg, key)
                        return got
                    except AttributeError:
                        continue

        try:
            raise KeyError(key_in)
        except Exception as e:
            raise

    @classmethod
    def factory(cls, paramstyle, case_insensitive=False):
        """
        return a Binder subclass instance appropriate
        to the underlying db library paramstyle bind variable

        :param paramstyle: parameter style string as per PEP-249
        :case_insensitive: %(custid)s will match {"custid":1} or {"CUSTID":2}, with priority
        going to the initial case.  mixed-case keys (custId) will only match {"custId":3}

        """

        try:
            inst = cls._di_paramstyle[paramstyle]()
            if case_insensitive:
                inst._key_expand = inst._case_insensitive

            return inst
        except KeyError:
            msg = """got:%s,
                  but expecting one of %s.
                  See
                  https://www.python.org/dev/peps/pep-0249/#paramstyle
                  for details""" % (
                paramstyle,
                "/".join(list(cls._di_paramstyle.keys())),
            )
            raise ValueError(msg)
        except NotImplementedError:
            msg = "%s is not implemented yet" % (paramstyle)
            raise NotImplementedError(msg)

    _di_paramstyle = {}

    # the regular expression pattern that looks for list type binds
    re_pattern_listsubstition = re.compile("%\([a-zA-Z0-9_]+\)l")

    # leading '__' variable name makes name clashes more unlikely
    T_LIST_KEYNAME = "__%s_%03d"

    # def _pre_process(self):
    #     """do nothing for now - intended to support list substitutions"""
    #     pass

    def _pre_process(self):
        li_listsubstition = self.re_pattern_listsubstition.findall(self.tqry)
        if li_listsubstition:
            self.preprocess_listsubstitution(li_listsubstition)

    def preprocess_listsubstitution(self, li_hit):
        """ this will transform %(xxx)l into %(__xxx_000)s, %(__xxx_001)s """

        di_list_sub = {}

        self.li_arg.insert(0, di_list_sub)

        for hit in li_hit:
            key = hit[2:-2]
            got = self._get_from_args(key)

            if not isinstance(got, (list, set)):
                raise ValueError(
                    "list substitutions require an iterable parameter: `%s` was of type `%s`"
                    % (key, type(got))
                )
                #
                # self.tqry = self.tqry.replace(hit, hit[:-1] + "s")
            else:

                li = []
                if not got:
                    # empty list or set
                    self.tqry = self.tqry.replace(hit, "NULL")
                    continue

                for ix, val in enumerate(got):
                    ikeyname = self.T_LIST_KEYNAME % (key, ix)
                    ikeyname_sub = "%%(%s)s" % (ikeyname)
                    di_list_sub[ikeyname] = val
                    li.append(ikeyname_sub)

                # replace the original bind %(xxx)l with
                # %(__xxx_000)s, %(__xxx_001)s, ...
                repval = ", ".join(li)
                self.tqry = self.tqry.replace(hit, repval)


class Binder_pyformat(Binder):
    """support Postgresql
       query template and substitution management for postgresql
       query is unchanged because postgresql is happy
       with %(somevar)s as a  bind
    """

    paramstyle = "pyformat"
    supports = "Postgresql"

    def _pre_process(self):
        li_listsubstition = self.re_pattern_listsubstition.findall(self.tqry)
        if li_listsubstition:
            self.preprocess_listsubstitution(li_listsubstition)

    def format(self, tqry, *args):
        """
        looks up substitutions and sets them up in dictionary self.sub

        postgresql accepts Python named variable so keeping the query as is

        select * from foo where bar = %(somebar)s"
        =>
        select * from foo where bar = %(somebar)s
        {"somebar" : value-found-for-somebar}
        """

        self.sub = {}
        self.li_arg = list(args)
        self.tqry = tqry

        self._pre_process()

        try:
            self.tqry % (self)
        except (Exception,) as e:
            raise

        # Postgresql query format stays as %(foo)s
        # so we just return the original query
        # (which _pre_process may have altered)
        return self.tqry, self.sub

    __call__ = format

    def __getitem__(self, key):
        if key in self.sub:
            return None

        got = self._get_from_args(key)
        self.sub[key] = got
        return None


PARAMSTYLE_QMARK = PARAMSTYLE_SQLITE = PARAMSTYLE_SQLSERVER = "qmark"


class BinderQmark(Binder):
    """ supports:  sqlite3, SQL Server
        query template and substitution management for sqlite3
        query changes from %(somevar)s to ?

        select * from foo where bar = %(somebar)s
        =>
        select * from foo where bar = ?,
        (value-found-for-somebar,)
    """

    paramstyle = PARAMSTYLE_QMARK
    supports = "sqlite3, mssql"
    qry_replace = "?"

    def format(self, tqry, *args):
        """
        looks up substitutions and sets them up in self.sub

        Note:
            Assuming both will be happy with a tuple.
            Might be one SQL Server needs a list instead.

        """
        self.tqry = tqry
        self._di_sub = {}
        self.sub = []
        self.li_arg = list(args)

        self._pre_process()

        try:
            qry = self.tqry % (self)
        except (Exception,) as e:
            raise

        return qry, tuple(self.sub)

    __call__ = format

    def __getitem__(self, key):
        """
        finds a substitution and append it to the bind list
        but also transforms the variable in the query to ?
        """

        qry_replace = self.qry_replace

        try:
            got = self._di_sub[key]
        except KeyError:
            got = self._di_sub[key] = self._get_from_args(key)

        self.sub.append(got)
        return qry_replace


class BinderFormat(BinderQmark):

    """supports: MySQL
       query template and substitution management for MySQL
       query changes from %(somevar)s to %s format
       parameters are (<var1>,<var2>,)

       Note: pretty much identical to BinderQmark/sqlite3
       except for the placeholder being %s

    """

    paramstyle = "format"
    supports = "MySQL"
    qry_replace = "%s"


class BinderNamed(Binder):
    """supports: Oracle
       query template and substitution management for Oracle
       query changes from %(somevar)s to :somevar format
       list-based substitutions:
           %(somelist)l :__somelist_000, :__somelist_001...
    """

    paramstyle = "named"
    supports = "Oracle"
    t_qry_replace = ":%s"

    def format(self, tqry, *args):
        """
        looks up substitutions and sets them up in self.sub

        but also transforms the query to Oracle named
        format
        "select * from foo where bar = %(somebar)s"
        =>
        "select * from foo where bar = :somebar "
        {"somebar" : value-found-for-somebar}
        
        """

        self.sub = {}
        self.li_arg = list(args)
        self.tqry = tqry

        self._pre_process()

        try:
            qry = self.tqry % (self)
        except (Exception,) as e:
            raise

        return qry, self.sub

    __call__ = format

    def __getitem__(self, key):
        """
        finds a substitution
        but also transforms the variable in the query to Oracle named
        format :foo
        """

        # already seen so already in the substition dict
        # replace the query's %(foo)s with :foo
        if key in self.sub:
            return self.t_qry_replace % (key)

        got = self._get_from_args(key)
        self.sub[key] = got
        return self.t_qry_replace % (key)

    """
    https://www.python.org/dev/peps/pep-0249/#paramstyle

    paramstyle  Meaning

    qmark       Question mark style, e.g. ...WHERE name=?   sequence
    numeric     Numeric, positional style, e.g. ...WHERE name=:1
    named       Named style, e.g. ...WHERE name=:name
    format      ANSI C printf format codes, e.g. ...WHERE name=%s
    pyformat    Python extended format codes, e.g. ...WHERE name=%(name)s
    """


ExperimentalBinderNamed = BinderNamed


class Binder_NotImplementedError(Binder):
    """not implemented yet"""

    paramstyle = "not implemented"

    def __init__(self, *args, **kwds):
        raise NotImplementedError()


# This is what decides how the Binder
# will process incoming template substitutions
Binder._di_paramstyle["pyformat"] = Binder_pyformat
Binder._di_paramstyle["named"] = BinderNamed
Binder._di_paramstyle[PARAMSTYLE_QMARK] = BinderQmark
Binder._di_paramstyle["format"] = BinderFormat

Binder._di_paramstyle["experimentalnamed"] = ExperimentalBinderNamed

# and these are not done yet
Binder._di_paramstyle["numeric"] = Binder_NotImplementedError



class SlotProxy(object):
    """allow assignment to instances originally using __slots__"""

    def __init__(self, obj):
        self.__dict__["_obj"] = obj

    def __setattr__(self, attrname, value):
        if attrname in self._obj.__slots__:
            setattr(self._obj, attrname, value)
        else:
            self.__dict__[attrname] = value

    def __getattr__(self, attrname):
        try:
            return getattr(self._obj, attrname)
        except AttributeError:
            raise


def rename_bindnames(tqry, li_adjust):
    """use this to alter the query template to match expected attribute names in bind objects/dictionaries

       For example, a predefined query may be:  "select * from customers where custid = %(custid)s"
       But you are repeatedly passing bind dictionaries like {"customer" : "cust001"}, {"customer" : "cust002"}

       in that case qry_template = rename_bindnames(qry_template, [("custid","customer")])
       can make your client code simpler and speed it up as well.

    """
    for bindname, attrname in li_adjust:
        from_ = "%(" + bindname + ")s"
        to_ = "%(" + attrname + ")s"
        tqry = tqry.replace(from_, to_)
    return tqry




###################################################################
# Python 2 to 3.  !!!TODO!!!p4- Simplify after Python support ends.
###################################################################
try:
    _ = basestring
except (NameError,) as e:
    basestring = str
###################################################################


from pynoorm.binder import Binder_pyformat


class PGDebugBinder(Binder_pyformat):

    name_debug_dict = "di_query_debug"

    def __init__(self, name_debug_dict=None, multiple=False):

        if name_debug_dict:
            self.name_debug_dict = name_debug_dict

        self.multiple = multiple


    def track(self, target, query, queryname, multiple=False):
        
        try:
            di_qry = getattr(target, self.name_debug_dict, {})

            multiple = multiple or self.multiple

            if multiple:
                li = di_qry.setdefault(queryname, [])
                li.append(query)
            else:
                di_qry[queryname] = query

        except (Exception,) as e:
            raise


    def __getitem__(self, key):
        if key in self.sub:
            return None

        got = self._get_from_args(key)

        if isinstance(got, basestring):
            got = "'%s'" % (got)
        elif got is None:
            got = "null"
        #!!!todo!!!p4 - handle date/datetimes...

        self.sub[key] = got
        return None

    @classmethod
    def get_query(cls, tqry, *args, **kwds):

        flaginsecure = kwds.get("flaginsecure", True)

        inst = cls()
        dtqry, dsub = inst.format(tqry, *args)
        return "insecure:\n" * flaginsecure + dtqry % dsub



from operator import attrgetter, itemgetter, setitem
import collections

from six import string_types

from .utils import SlotProxy


########### debugging aids ##################
import pdb
from traceback import print_exc as xp


def cpdb():
    """do-nothing debugger test"""
    return cpdb.enabled


cpdb.enabled = False
########### debugging aids ##################


class LinkResultHelper(object):
    """returned by `Linker.link` and can be used to see what wasn't linked"""

    def __init__(_this, **kwds):
        """the odd bit with `_this` as first paremeter is because there is a `self`
           in the kwds from Linker.link, so that would clash.
           Assign all the Linker.link parameters to the instance's __dict__, then
           we go back to standard self"""

        # `self` in the kwds refers to the Linker instance
        kwds["linker"] = kwds.pop("self")
        # store everything that was passed as parameters to `link`.
        _this.__dict__.update(**kwds)

        self = _this
        self.right_orphans = []
        self.left_orphans = []
        self.exception = None

    def add_right_orphan(self, o):
        """track that linker.link didn't find a di_left[right.keyval]"""
        self.right_orphans.append(o)

    def initialize_rights(self):
        """initialize right-side objects that didn't get linked"""
        try:
            li = self.right_orphans
            return self._initialize(li, self.attrname_on_right, self.type_on_right)
        except (Exception,) as e:  # pragma: no cover
            if cpdb():
                pdb.set_trace()
            raise

    def _initialize(self, li, attrname, type_):
        """initialize objects that didn't get linked"""
        if not li:  # pragma: no cover
            return self

        linker = self.linker
        getter = self.linker._get_getter(li[0], attrname)
        _empty_setter = None
        for obj in li:
            try:
                v = getter(obj)
            except (AttributeError, KeyError):
                _empty_setter = _empty_setter or linker._get_empty_setter(
                    obj, attrname, type_
                )
                _empty_setter(obj, attrname, type_)
        return self

    def initialize_lefts(self):
        """initialize left-side objects that didn't get linked"""
        try:
            li = list(self.left.values())
            return self._initialize(li, self.attrname_on_left, self.type_on_left)
        except (Exception,) as e:  # pragma: no cover
            if cpdb():
                pdb.set_trace()
            raise


class Linker(object):
    """
        Used to set up very fast one or two-way links between
        objects.  You can think 'parent' for left and 'child' for right
        but that's not strictly true as sibling and scalar relationships
        are supported.  many-to-many is not supported.

    """

    TYPE_SCALAR = None

    def __init__(self, key_left):
        """key_left is either a string or tuple of strings
           stating which attributes/keys on future objects
           identify them.
           ex:  ("custid","order_id") for CustomerOrder Table
        """
        self.key_left = key_left

    def __repr__(self):
        return "%s on fields: %s" % (self.__class__.__name__, self.key_left)

    def _get_getter(self, obj, key):
        """returns a getter function appropriate for the getitem/getattr support in `obj`"""
        try:
            if isinstance(obj, collections.Mapping):
                if isinstance(key, string_types):
                    return itemgetter(key)
                elif isinstance(key, collections.Sequence):
                    return itemgetter(*key)
                else:
                    raise TypeError(
                        "expecting a string or tuple of strings as key.  got:%s[%s]"
                        % (str(key), type(key))
                    )
            else:
                if isinstance(key, string_types):
                    return attrgetter(key)
                elif isinstance(key, collections.Sequence):
                    return attrgetter(*key)
                else:
                    raise TypeError(
                        "expecting a string or tuple of strings as key.  got:%s[%s]"
                        % (str(key), type(key))
                    )

        except (Exception,) as e:  # pragma: no cover
            if cpdb():
                pdb.set_trace()
            raise

    def dict_from_list(self, li):
        """prepare a dictionary for linking based on `left_key`
           :param li:  a list of objects or dictionaries
        """

        key_left = self.key_left

        di_left = dict()

        li2 = None
        if isinstance(li, dict):
            li2 = li.values()
        li2 = li2 or li

        get_key = None
        for o_left in li2:
            get_key = get_key or self._get_getter(o_left, key_left)
            try:
                keyval = get_key(o_left)
                # NOTE:  at this point, if we used a list instead of a simple assignment could we do m-n?
                di_left[keyval] = o_left
            except (Exception,) as e:  # pragma: no cover
                if cpdb():
                    pdb.set_trace()
                raise

        return di_left

    def link(
        self,
        left,
        right,
        attrname_on_left,
        setter_left=None,
        type_on_left=list,
        dictkey_attrname_left=None,
        key_right=None,
        setter_right=None,
        attrname_on_right=None,
        type_on_right=None,
    ):
        """
        :param left: a dictionary of objects or dictionaries which will be linked to right-side objects
        :param right: a list(iterator?) of objects or dictionaries.  you can also pass in a dictionary, its values will be used in that case
        :param attrname_on_left: the attribute name (or dictionary key) where the right-side object ref will be stored
        :param setter_left:  you can pass a callback to assign the right-side to left-side yourself.
                             call signature:  f(o_target, attrname, o_link)
        :param type_on_left:  None/Linker.scalar - direct assignment o_left.attrname_on_left = o_right
                              list (the default) - append each right-side object
                              dict - references are stored in a dict, but that requires dictkey_attrname_left to have been set as well.
                              
        :param dictkey_attrname_left: if your target's attribute is a dictonary, you need to provide the field that will be used on for that key
                                      ex:  attrname_on_left="tags", type_on_left=dict, dictkey_attrname_left="tagtype"
        :param key_right: specifying something here allows you to alias fields used in key_left
        :param setter_right: see setter_left
        :param attrname_on_right:  passing a value means a 2-way link
        :param type_on_right:  scalar is assumed, but list is also supported

        :return: LinkResultHelper instance to check orphans/assist initializations when needed

        """

        try:

            self.helper = LinkResultHelper(**locals())

            try:
                assert isinstance(attrname_on_left, string_types)
            except (AssertionError,) as e:  # pragma: no cover
                raise TypeError(
                    "attrname_on_left needs to be a valid python variable name"
                )

            key_left = self.key_left
            key_right = key_right or key_left

            # grab some sample objects from left and right
            # and use the samples to figure out getters and setters
            try:
                sample_left = next(iter(left.values()))
            except (StopIteration,) as e:
                self.helper.exception = ValueError("empty left", e)
                return self.helper

            it_right = iter(right)
            try:
                sample_right = next(it_right)
            except (StopIteration,) as e:
                self.helper.exception = ValueError("empty right", e)
                return self.helper

            get_key = self._get_getter(sample_right, key_right)
            setter_left = setter_left or self._get_setter(
                sample_left,
                attrname_on_left,
                type_on_left,
                dictkey_attrname_left,
                sample_right,
            )

            if attrname_on_right:
                setter_right = setter_right or self._get_setter(
                    sample_right, attrname_on_right, type_on_right
                )
            else:
                setter_right = None

            # do we need to set left-only? or left and right?
            if setter_right:
                prepped = self._preppedlinkleftright
            else:
                prepped = self._preppedlinkleft

            # finally, call the actual link, on the sample, then the rest
            for right_ in [[sample_right], it_right]:

                prepped(
                    left=left,
                    right=right_,
                    attrname_on_left=attrname_on_left,
                    setter_left=setter_left,
                    key_left=key_left,
                    key_right=key_right,
                    setter_right=setter_right,
                    attrname_on_right=attrname_on_right,
                    get_key=get_key,
                )

        except (Exception,) as e:  # pragma: no cover
            if cpdb():
                pdb.set_trace()
            raise
        else:
            return self.helper

    def _preppedlinkleft(
        self,
        left,
        right,
        attrname_on_left,
        setter_left,
        key_left,
        key_right,
        setter_right,
        attrname_on_right,
        get_key,
    ):
        """called from link, with all the getters/setters pre-discovered
           and knows that it won't be setting anything on the right side.
           however the signature is the same as `_preppedlinkleftright`, to unify and simplify
           calling
        """

        helper = self.helper
        for o_right in right:
            keyval = get_key(o_right)
            o_left = left.get(keyval, None)
            if o_left is None:
                helper.add_right_orphan(o_right)
                continue
            setter_left(o_left, attrname_on_left, o_right)

    def _preppedlinkleftright(
        self,
        left,
        right,
        attrname_on_left,
        setter_left,
        key_left,
        key_right,
        setter_right,
        attrname_on_right,
        get_key,
    ):
        """see `_preppedlinkleft`, but this will set attributes on the right
        """

        helper = self.helper
        for o_right in right:
            keyval = get_key(o_right)
            o_left = left.get(keyval, None)
            if o_left is None:
                helper.add_right_orphan(o_right)
                continue
            setter_left(o_left, attrname_on_left, o_right)
            setter_right(o_right, attrname_on_right, o_left)

    def _get_empty_setter(self, obj, attrname_on_tgt, type_on_tgt):
        """initialize the attribute to an appropriate empty value"""
        try:
            if isinstance(obj, collections.Mapping):

                def setdefault(tgt, attrname, value):
                    tgt.setdefault(attrname, value())

                def setvalue(tgt, attrname, value):
                    tgt[attrname] = value

                if callable(type_on_tgt):
                    return setdefault
                elif type_on_tgt is None:
                    return setvalue
                else:
                    raise NotImplementedError()
            else:

                def setdefault(tgt, attrname, value):
                    setattr(tgt, attrname, value())

                def setvalue(tgt, attrname, value):
                    setattr(tgt, attrname, value)

                if callable(type_on_tgt):
                    return setdefault
                elif type_on_tgt is None:
                    return setvalue
                else:
                    raise NotImplementedError()

        except (Exception,) as e:  # pragma: no cover
            if cpdb():
                pdb.set_trace()
            raise

    supported_target_types = [list, None]

    def _get_setter(
        self, obj, attrname_on_tgt, type_on_tgt, dictkey_attrname=None, o_src=None
    ):
        """determines the appropriate function to set values
           - mapping types will privilege setitem
           - other instance will use setattr

           type_on_tgt (default assumption is left 0..1 <=> 0..N right meaning that the left attribute is a list)
           also determines how to initialize attribute and add values.
        """

        assert isinstance(attrname_on_tgt, string_types)
        try:
            if isinstance(obj, collections.Mapping):

                def append(tgt, attrname, value):
                    li = tgt.setdefault(attrname, [])
                    li.append(value)

                if type_on_tgt == list:
                    return append
                elif type_on_tgt == dict:
                    raise NotImplementedError()
                elif type_on_tgt is None:
                    return setitem
                else:
                    raise TypeError(
                        "unsupported target type:%s.  Supported are: %s"
                        % (type_on_tgt, self.supported_target_types)
                    )
            else:

                def append(tgt, attrname, value):
                    li = getattr(tgt, attrname, None)
                    if li is None:
                        li = []
                        setattr(tgt, attrname, li)
                    li.append(value)

                if type_on_tgt == list:
                    return append
                elif type_on_tgt == dict:

                    getter_left = self._get_getter(o_src, dictkey_attrname)

                    def setdict(tgt, attrname, value):

                        di = getattr(tgt, attrname, None)
                        if di is None:
                            di = {}
                            setattr(tgt, attrname, di)
                        keyval = getter_left(value)
                        di[keyval] = value

                    return setdict

                elif type_on_tgt is None:
                    return setattr
                else:
                    raise TypeError(
                        "unsupported target type:%s.  Supported are: %s"
                        % (type_on_tgt, self.supported_target_types)
                    )

        except (Exception,) as e:  # pragma: no cover
            if cpdb():
                pdb.set_trace()
            raise



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import sys
import os
import re


if sys.version_info < (2, 7):
    raise Exception("PyNoORM requires Python 2.7 or higher.")

if sys.version_info >= (3,):
    if sys.version_info < (3, 3):
        raise Exception("PyNoORM requires Python 3.3 or higher")


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


# with open('README.rst') as readme_file:
#     readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    # TODO: put package requirements here
]

test_requirements = [
    # TODO: put package test requirements here
]


def parse_readme(text):
    """start on reStructuredText banner and end at software declaration"""

    start = re.compile("~~~~~~~", re.IGNORECASE)
    end = re.compile("Free software:", re.IGNORECASE)
    from_ = to_ = description = None

    lines = text.split("\n")

    for lineno, line in enumerate(lines):

        if from_ is None and start.search(line):
            from_ = lineno - 1
            description = lines[from_].strip()

        if to_ is None and end.search(line):
            to_ = lineno

    return description, "\n".join(lines[from_:to_])


with open(os.path.join(os.path.dirname(__file__), "README.rst")) as r_file:
    description, readme = parse_readme(r_file.read())

    assert description.strip()
    assert readme.strip()


setup(
    name="pynoorm",
    version="0.4.3",
    description=description,
    long_description=readme + "\n\n" + history,
    author="JL Peyret",
    author_email="jpeyret@gmail.com",
    url="https://github.com/jpeyret/pynoorm",
    packages=["pynoorm"],
    package_dir={"pynoorm": "pynoorm"},
    include_package_data=True,
    install_requires=requirements,
    license="MIT License",
    zip_safe=False,
    keywords="sql database multiplatform",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Database :: Front-Ends",
        "Topic :: Utilities",
        "Operating System :: OS Independent",
    ],
    test_suite="tests",
    tests_require=test_requirements,
)



########### debugging aids ##################
import pdb
from traceback import print_exc as xp


def cpdb():
    """do-nothing debugger test"""
    return cpdb.enabled


cpdb.enabled = False
########### debugging aids ##################


class AsObject(object):
    """used to convert result to objects"""

    def __init__(self, **kwds):
        self.__dict__.update(**kwds)


class Overrides(object):
    def __init__(self, attrnames, type_right=object, to_object_class=None):
        """prepare a dictionary for linking based on `left_key`
           :param attrnames:  attributes to set on overrides.  several forms are supported
           "tax" - a string, will get `tax` the right and assign it to `tax`
           ("rate","tax") assigns `tax` to `rate` as an alias.
           ["tax","state"] will get and assign both tax and state.
           Note that all attributes are expected to exist on the sources, KeyError or AttributeError
           happen otherwise.
           :param type_right: by default, attribute retrieval are done by `getattr` because objects
           are assumed.  passing in `dict` will use `dict[key]` instead.
           :param to_object_class:  used by `as_objects`.  You can pass in a class but it needs to 
           support keyword assignment for all attributes' assign aliases.

        """

        self.attrnames = attrnames
        self._dict = {}

        if to_object_class is None:
            self.to_object_class = AsObject
        else:
            self.to_object_class = to_object_class

        if issubclass(type_right, dict):
            self.setter = self._setter_attr_from_dict
        else:
            self.setter = self._setter_attr_from_object

        self._mapping = []

        if not isinstance(attrnames, list):
            attrnames = [attrnames]

        for attrname in attrnames:
            if isinstance(attrname, tuple):
                attrname_l, attrname_r = attrname
            else:
                attrname_l = attrname_r = attrname
            self._mapping.append((attrname_l, attrname_r))

    def __getitem__(self, attrname):
        return self._dict[attrname]

    def values(self):
        """needed to tell the Linker to expect dictionaries"""
        return [{}]

    def get(self, key, default=None):
        """returns an empty dictionary if the key doesnt exist yet"""

        try:
            return self._dict.setdefault(key, {})
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def _setter_attr_from_dict(self, o_left, attrnames, o_right):
        """assigns right-side values to the working dictonary's aliases """
        try:
            for attrname_l, attrname_r in self._mapping:
                o_left[attrname_l] = o_right[attrname_r]
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def _setter_attr_from_object(self, o_left, attrname, o_right):
        """assigns right-side attributes to the working dictonary's aliases """
        try:
            for attrname_l, attrname_r in self._mapping:
                o_left[attrname_l] = getattr(o_right, attrname_r)
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def as_dict(self):
        """return overrides as dictionaries"""
        try:
            return self._dict.copy()
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def as_objects(self):
        """return overrides as objects, using `to_object_class`, if provided """
        try:
            di = {}
            cls_ = self.to_object_class
            for k, v in self._dict.items():
                di[k] = cls_(**v)
            return di
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def link(self, linker, data):
        """calls the linker providing itself as the lookup dictionary"""
        try:
            linker.link(self, data, attrname_on_left="_", setter_left=self.setter)
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise
